#include <assert.h>
#include <X11/Xlib.h>
#include <stdio.h>
#include <stdlib.h>

#include "ttf.h"
#include "pixmap.h"

/***********************************************************************
  Debugging
*/

#define DEBUG_MISC 1

#ifndef DEBUG
#define DEBUG 0
#endif

/***********************************************************************
  Useful macros
*/

/* Remove any old declaration */
#ifdef ABS
#undef ABS
#endif

#define ABS(x) ((x) >= 0 ? (x) : -(x))

/***********************************************************************

   Determine when to stop subdividing

   This one is quite simpleminded, it just checks to see how far apart
   the on-curve points are.

   A better one might be to check if the off-curve point falls on a
   straight line between the on-curve points.

*/

#define TERM(xA, yA, xB, yB, xC, yC) \
	(ABS(xA-xC) <= termval && ABS(yA-yC) <= termval)

/************************************************************************

  ttfBezier - Subdivide a Cubic Bezier into line segments

  XPoint *pt		Where to store the generated points
  return XPoint * 	Pointer to element after last point stored
  int depth		The deepest to go  (to make sure we finish)
  long xA, xB, yC, yC	on-curve points
  long xB, yB		off-curve control point

*/

XPoint *ttfBezier(XPoint * pt,
		  int depth,
		  long xA, long yA,
		  long xB, long yB,
		  long xC, long yC,
		  long termval
){
    long xAB, yAB;
    long xBC, yBC;
    long xABC, yABC;

    /* Time to quit? */
    if (!depth || TERM(xA, yA, xB, yB, xC, yC)) {
	/* Since the order of subdivision has been carefully chosen, the
       points will be added to the list in the correct order */
	pt->x = xC;
	pt->y = yC;
	pt++;
	return pt;
    }
    /* If all coordinate values can be guaranteed to be positive, all
     divisions can be replaced by shifts */

    xAB = (xA + xB) / 2;	/* AB = midpoint(A, B) */
    yAB = (yA + yB) / 2;

    xBC = (xB + xC) / 2;	/* BC = midpoint(B, C) */
    yBC = (yB + yC) / 2;

    xABC = (xAB + xBC) / 2;	/* ABC = midpoint(AB, BC) */
    yABC = (yAB + yBC) / 2;

    pt = ttfBezier(pt,
		   depth - 1,
		   xA, yA,
		   xAB, yAB,
		   xABC, yABC,
		   termval);

    pt = ttfBezier(pt,
		   depth - 1,
		   xABC, yABC,
		   xBC, yBC,
		   xC, yC,
		   termval);

    return pt;
}

/***********************************************************************

  ttfLinearize - convert a TrueType glyph to a polygon

  XPoint *lines		Where to store the coordinates of the line segments
  return XPoint *lines	Pointer to element after last written
  int nPoints		No of points defining the glyph
  ttfPoint *points	Array of points

*/

XPoint *ttfLinearize(
    XPoint * lines,
    int nPoints,
    ttfPoint * points,
    int ppm,
    int upm)
{
    BOOL lastWasOnPoint;
    XPoint onPoint, offPoint;
    XPoint *pl;			/* Output line-segments */
    ttfPoint *pp;		/* Input points defining glyph */
    int termval;

#if DEBUG & DEBUG_MISC
    fprintf(stderr, "Linearizing glyph\n");
#endif

    termval = upm / ppm;
    if (termval < 1) termval = 1;
    pl = lines;
    pp = points;

    lastWasOnPoint = 0;

    if (!points[0].flags & TTFPOINT_ON)
	fprintf(stderr, "WARNING! First point of curve was NOT an on point\n");

    /* Initialize with first point which must be an on point */
    pl->x = pp->x;
    pl->y = pp->y;
    pl++;

    onPoint.x = pp->x;
    onPoint.y = pp->y;
    lastWasOnPoint = 1;
    pp++;
    nPoints--;

    while (nPoints--) {
	if (pp->flags & TTFPOINT_ON) {
	    /* This is an on-curve point */

	    if (lastWasOnPoint) {
		/* Two on-curve points in a row constitutes a straight line */
		pl->x = pp->x;
		pl->y = pp->y;
		pl++;
	    } else {
		/* We have a complete bezier, draw it */
		pl = ttfBezier(pl, DEPTH,
			       onPoint.x, onPoint.y,
			       offPoint.x, offPoint.y,
			       pp->x,
			       pp->y,
			       termval);
	    }

	    /* Save the on point for later */
	    onPoint.x = pp->x;
	    onPoint.y = pp->y;
	    lastWasOnPoint = 1;
	} else {
	    /* This is an off-curve point */

	    if (!lastWasOnPoint) {
		/* If the last point wasn't an on-curve point, we have to
		     insert one halfway between the last and current off-curve
		     points.
		     
		     I don't think this is correct -- hjp
		 */

		int x, y;

		/* Create a new on-curve point */
		x = (offPoint.x + pp->x) / 2;
		y = (offPoint.y + pp->y) / 2;

		/* Draw the bezier */
		pl = ttfBezier(pl, DEPTH,
			       onPoint.x, onPoint.y,
			       offPoint.x, offPoint.y,
			       x,
			       y,
			       termval);

		/* Save this on-curve point for later */
		onPoint.x = x;
		onPoint.y = y;
	    }
	    /* Save the off curve point for later */
	    offPoint.x = pp->x;
	    offPoint.y = pp->y;
	    lastWasOnPoint = 0;
	}

	/* Next control point please */
	pp++;
    }

    /* Close the curve */
    if (!lastWasOnPoint) {
	pl = ttfBezier(pl, DEPTH,
		       onPoint.x, onPoint.y,
		       offPoint.x, offPoint.y,
		       points->x,
		       points->y,
		       termval);
    }
    pl->x = points->x;
    pl->y = points->y;
    pl++;

#if DEBUG & DEBUG_MISC
    fprintf(stderr, "  No of line segments: %d\n", pl - lines);
#endif

    return pl;
}

/***********************************************************************
  This is really ugly code.  I got quite tired, so I didn't bother
  with dynamic allocation of memory or anything fancy.  It should be
  possible to replace this with X's rasterizer.

  What it does is to take a polygon and scan-convert it by walking
  through the list of line segments, and for each line segment
  calculate which scanlines it crosses and where.

  To avoid setting duplicate on or off points we remember the last
  point set. If the current point is exactly the same as the last
  one, we don't set it.

  The output is a Render List where each line has a list of crossings
  and wether the winding number should increase or decrease.

  Take a look at Graphics Gems I, `Fast Scan Conversion of Arbitrary
  Polygons' for more theory on how it works.

*/

static void addToList(
    ttfRenderList * renderList,
    int x, int y, int on,
    int ppm, int upm
){
#if 1
    y = y * ppm / upm;
#endif
    assert(0 <= y && y < 1024);
    renderList[y].columns[renderList[y].nColumns] = x;
    renderList[y].on[renderList[y].nColumns] = on;
    renderList[y].nColumns++;
}

void ttfRender(
    ttfRenderList * renderList,
    int nLines,
    XPoint * lines,
    int ppm,
    int upm)
{
    int i;
    int bias = upm / (2 * ppm);
    int lx = INT_MIN, ly = INT_MIN, lo = INT_MIN;

    for (i = 1; i < nLines; i++) {
	int x, y;
	int dx, dy;

	dx = lines[i].x - lines[i - 1].x;
	dy = lines[i].y - lines[i - 1].y;

	if (dy > 0) {
	    int py = (lines[i - 1].y - bias) * ppm / upm;

	    for (;;) {
	        int y1;
		y = py * upm / ppm + bias;
		y1 = (py - 1) * upm / ppm + bias;
		if (y < lines[i - 1].y) {
		    fprintf(stderr,
		    	    "adjusting: y = %d, y1 = %d, wanted = %d\n",
		    	    y, y1, lines[i - 1].y);
		    py ++;
		} else if (y1 >= lines[i - 1].y) {
		    fprintf(stderr,
		    	    "adjusting: y = %d, y1 = %d, wanted = %d\n",
		    	    y, y1, lines[i - 1].y);
		    py --;
		} else {
		    break;
		}
	    }
	    /* Moving down, so the interior is to the left */
	    for (;
	    	 y <= lines[i].y;
		 py++, y = py * upm / ppm + bias) {
		x = lines[i - 1].x + dx * (y - lines[i - 1].y) / dy;
		if (/* x != lx || */ y != ly || False != lo) {
		    addToList(renderList, x, y, False, ppm, upm);	/* add an off point */
		    lx = x; ly = y; lo = False;
		    fprintf(stderr, "renderList: %d %d %d\n", lx, ly, lo);
		}
	    }
	} else if (dy < 0) {
	    int py = (lines[i - 1].y + bias) * ppm / upm;

	    for (;;) {
	        int y1;
		y = py * upm / ppm + bias;
		y1 = (py + 1) * upm / ppm + bias;
		if (y > lines[i - 1].y) {
		    fprintf(stderr,
		    	    "adjusting: y = %d, y1 = %d, wanted = %d\n",
		    	    y, y1, lines[i - 1].y);
		    py --;
		} else if (y1 <= lines[i - 1].y) {
		    fprintf(stderr,
		    	    "adjusting: y = %d, y1 = %d, wanted = %d\n",
		    	    y, y1, lines[i - 1].y);
		    py ++;
		} else {
		    break;
		}
	    }
	    /* Moving up, so the interior is to the right */
	    for (;
		 y >= lines[i].y;
		 py--, y = py * upm / ppm + bias) {
		x = lines[i - 1].x + dx * (y - lines[i - 1].y) / dy;
		if (/* x != lx || */ y != ly || True != lo) {
		    addToList(renderList, x, y, True, ppm, upm);	/* add an on point */
		    lx = x; ly = y; lo = True;
		    fprintf(stderr, "renderList: %d %d %d\n", lx, ly, lo);
		}
	    }
	} else {
	    /* Horizontal lines are a special case */
	    /* just ignore for now */
	}
    }
}

/***********************************************************************
  Generate a bitmap from the render list.

  This is a quite straightforward function.  Sort the coordinates of
  each line, and start drawing from the left edge.  Start turning on
  pixels after an on-crossing and turning off after an off-crossing.

*/

static int compare(ttfRenderList * rl, int a, int b)
{
    if (rl->columns[a] < rl->columns[b])
	return 1;
    if (rl->columns[a] > rl->columns[b])
	return 0;
    if (rl->on[a] > rl->on[b])
	return 1;
    return 0;
}

void ttfRasterize(
    pixmapT * bitmap,
    int xoff,
    ttfRenderList * renderList,
    int ppm, int upm
){
    int i, j, k;
    int rows = bitmap->rows;
    int cols = bitmap->cols;
    int bias = upm / (2 * ppm);

    for (i = 0; i < rows; i++) {
	int on;
	int x;

	for (j = 1; j < renderList[i].nColumns; j++)
	    for (k = 0; k < j; k++)
		if (compare(&renderList[i], j, k)) {
		    int col = renderList[i].columns[j];
		    int o = renderList[i].on[j];
		    renderList[i].columns[j] = renderList[i].columns[k];
		    renderList[i].on[j] = renderList[i].on[k];
		    renderList[i].columns[k] = col;
		    renderList[i].on[k] = o;
		}
	if (renderList[i].nColumns) {
	    on = 0;
	    x = 0;
	    for (j = 0; j < renderList[i].nColumns; j++) {
		while ((x * upm / ppm + bias) < renderList[i].columns[j]) {
		    if (on > 0) {
			if (x + xoff < cols) {
			    bitmap->pixels[i][x + xoff] = pixel(255, 0, 0, 0);
			}
		    }
		    x++;
		}

		if (renderList[i].on[j])
		    on++;
		else
		    on--;

		if (on < 0) {
		    int k;
		    fprintf(stderr, "BUG! on is negative\n");
		    fprintf(stderr, "\t%d (%d): ",
		    	i, i * upm / ppm + bias);
		    for (k = 0; k < renderList[i].nColumns; k++) {
			fprintf(stderr, " [%d:%d]",
				renderList[i].columns[k],
				renderList[i].on[k]);
		    }
		    fprintf(stderr, "\n");

		}
	    }
	}
    }

}
